#include "script.h"
#include "grub_script.tab.h"

struct grub_parser_state_transition
{
  /* The state that is looked up.  */
  grub_parser_state_t from_state;

  /* The next state, determined by FROM_STATE and INPUT.  */
  grub_parser_state_t to_state;

  /* The input that will determine the next state from FROM_STATE.  */
  char input;

  /* If set to 1, the input is valid and should be used.  */
  int keep_value;
};

/* All the possible state transitions on the command line.  If a
   transition can not be found, it is assumed that there is no
   transition and keep_value is assumed to be 1.  */
static struct grub_parser_state_transition state_transitions[] =
{
  { GRUB_PARSER_STATE_TEXT, GRUB_PARSER_STATE_QUOTE, '\'', 0},
  { GRUB_PARSER_STATE_TEXT, GRUB_PARSER_STATE_DQUOTE, '\"', 0},
  { GRUB_PARSER_STATE_TEXT, GRUB_PARSER_STATE_VAR, '$', 0},

  { GRUB_PARSER_STATE_QUOTE, GRUB_PARSER_STATE_TEXT, '\'', 0},

  { GRUB_PARSER_STATE_DQUOTE, GRUB_PARSER_STATE_TEXT, '\"', 0},
  { GRUB_PARSER_STATE_DQUOTE, GRUB_PARSER_STATE_QVAR, '$', 0},

  { GRUB_PARSER_STATE_VAR, GRUB_PARSER_STATE_VARNAME2, '{', 0},
  { GRUB_PARSER_STATE_VAR, GRUB_PARSER_STATE_VARNAME, 0, 1},
  { GRUB_PARSER_STATE_VARNAME2, GRUB_PARSER_STATE_TEXT, '}', 0},

  { GRUB_PARSER_STATE_QVAR, GRUB_PARSER_STATE_QVARNAME2, '{', 0},
  { GRUB_PARSER_STATE_QVAR, GRUB_PARSER_STATE_QVARNAME, 0, 1},
  { GRUB_PARSER_STATE_QVARNAME, GRUB_PARSER_STATE_DQUOTE, ' ', 1},
  { GRUB_PARSER_STATE_QVARNAME, GRUB_PARSER_STATE_TEXT, '\"', 0},
  { GRUB_PARSER_STATE_QVARNAME2, GRUB_PARSER_STATE_DQUOTE, '}', 0},

  { 0, 0, 0, 0}
};

/* Determines the state following STATE, determined by C.  */
grub_parser_state_t
grub_parser_cmdline_state (grub_parser_state_t state, char c, char *result)
{
  struct grub_parser_state_transition *transition;
  struct grub_parser_state_transition *next_match = 0;
  struct grub_parser_state_transition default_transition;
  int found = 0;

  default_transition.to_state = state;
  default_transition.keep_value = 1;

  /* Look for a good translation.  */
  for (transition = state_transitions; transition->from_state; transition++)
    {
      /* An exact match was found, use it.  */
      if (transition->from_state == state && transition->input == c)
	{
	  found = 1;
	  break;
	}

      /* A less perfect match was found, use this one if no exact
	 match can be found.  */
      if (transition->from_state == state && transition->input == 0)
	next_match = transition;
    }

  if (! found)
    {
      if (next_match)
	transition = next_match;
      else
	transition = &default_transition;
    }

  if (transition->keep_value)
    *result = c;
  else
    *result = 0;
  return transition->to_state;
}

void
grub_script_lexer_ref (struct grub_lexer_param *state)
{
  state->refs++;
}

void
grub_script_lexer_deref (struct grub_lexer_param *state)
{
  state->refs--;
}

#define GRUB_PARSER_BUFLEN	4096

struct grub_lexer_param *
grub_script_lexer_init (char *script, grub_err_t (*getline) (char **))
{
  struct grub_lexer_param *param;

  param = grub_malloc (sizeof (*param));
  if (! param)
    return 0;

  param->state = GRUB_PARSER_STATE_TEXT;
  param->read_ahead = ' ';
  param->getline = getline;
  param->refs = 0;
  param->done = 0;
  param->newscript = 0;
  param->script = script;
  param->buffer = grub_malloc(GRUB_PARSER_BUFLEN);
  if (! param->buffer)
    {
      grub_free (param);
      return 0;
    }
  param->offset = 0;

  return param;
}

char
grub_script_lexer_getch(struct grub_lexer_param *state)
{
  if (state->done)
    return 0;

  if (! state->script)
    {
      if (((state->refs) ||
           (state->state == GRUB_PARSER_STATE_QUOTE) ||
           (state->state == GRUB_PARSER_STATE_DQUOTE))
          && state->getline)
	{
          state->getline (&state->newscript);
          state->script = state->newscript;
	}
      if (! state->script)
        {
          state->done=1;
          return 0;
        }
    }

  if (! *state->script)
    {
      if (state->newscript)
        grub_free (state->newscript);
      state->newscript=state->script=0;
      return '\n';
    }
  else
    return *(state->script++);
}

#define is_blank(c)	(c==' ') || (c=='\t')
#define is_delimiter(c)	is_blank(c) || (c==';') || (c=='{') || (c=='}') || (c=='\n')
#define save_ch(state,ch)   if (state->offset < GRUB_PARSER_BUFLEN - 1) state->buffer[state->offset++]=ch;

int
grub_script_yylex (YYSTYPE *yylval, struct grub_parser_param *parsestate)
{
  grub_parser_state_t newstate;
  char use, ch;
  int is_text,res;
  struct grub_lexer_param *state = parsestate->lexerstate;
  struct grub_script_node *node = 0;

  void add_str(void)
    {
      struct grub_script_node *str;

      state->buffer[state->offset]=0;
      str=grub_script_newstr(parsestate,(is_text)?GRUB_SCRIPT_NODE_TYPE_STR:GRUB_SCRIPT_NODE_TYPE_VAR,state->buffer);
      if (! node)
        {
          node = grub_script_newnode(parsestate,GRUB_SCRIPT_NODE_TYPE_TEXT);
          node->child=grub_script_getnode(parsestate, str);
        }
      else
        grub_script_catnode(parsestate, node->child, str);

      state->offset = 0;
    }

  if (state->done)
    return 0;

  ch=state->read_ahead;
  if (state->state == GRUB_PARSER_STATE_TEXT)
    {
      while (is_blank(ch))
        {
          ch=grub_script_lexer_getch(state);
          if (state->done)
            return 0;
        }

      if (ch=='#')
        do
          {
            ch=grub_script_lexer_getch(state);
          } while (ch!='\n');

      if (is_delimiter(ch))
        {
          state->read_ahead=' ';
          return ch;
        }
    }

  is_text=1;
  state->offset=0;
  while (! state->done)
    {
      int new_is_text;

      if ((state->state == GRUB_PARSER_STATE_TEXT) && (is_delimiter(ch)))
        {
          newstate=state->state;
          break;
        }
      if ((ch=='\\') &&
          ((state->state == GRUB_PARSER_STATE_TEXT) ||
           (state->state == GRUB_PARSER_STATE_DQUOTE)))
        {
          grub_script_lexer_ref(state);
          ch=grub_script_lexer_getch(state);
          if (ch=='\n')
            ch=grub_script_lexer_getch(state);
          else
            {
              if ((state->state == GRUB_PARSER_STATE_DQUOTE) &&
                  (ch!='"') && (ch!='\\') && (ch!='$'))
                save_ch(state, '\\');
            }
          grub_script_lexer_deref(state);
          use=ch;
          newstate=state->state;
        }
      else
        newstate=grub_parser_cmdline_state (state->state, ch, &use);

      if (((newstate == GRUB_PARSER_STATE_VARNAME) ||
           (newstate == GRUB_PARSER_STATE_QVARNAME)) &&
          (is_delimiter(ch) || (ch=='/') || (ch=='$')))
        {
          if ((newstate == GRUB_PARSER_STATE_QVARNAME) ||
              (ch=='$') || (ch=='/'))
            new_is_text = 1;
          else
            break;
          if (ch=='$')
            use=0;
        }
      else
        new_is_text=(newstate<=GRUB_PARSER_STATE_TEXT);

      if (is_text!=new_is_text)
        {
          if (state->offset)
            add_str();
          is_text=new_is_text;
        }

      if (use)
        save_ch(state, use);

      state->state=newstate;
      ch=grub_script_lexer_getch(state);
    }

  if (newstate==GRUB_PARSER_STATE_VARNAME)
    newstate=GRUB_PARSER_STATE_TEXT;

  res=GRUB_PARSER_TOKEN_TEXT;
  if (newstate != GRUB_PARSER_STATE_TEXT)
    res=GRUB_PARSER_TOKEN_ERROR;
  else if ((is_text) && (! node))
    {
      state->buffer[state->offset]=0;
      if (! grub_strcmp (state->buffer, "while"))
        res=GRUB_PARSER_TOKEN_WHILE;
      else if (! grub_strcmp (state->buffer, "if"))
        res=GRUB_PARSER_TOKEN_IF;
      else if (! grub_strcmp (state->buffer, "function"))
        res=GRUB_PARSER_TOKEN_FUNCTION;
      else if (! grub_strcmp (state->buffer, "menuentry"))
        res=GRUB_PARSER_TOKEN_MENUENTRY;
      else if (! grub_strcmp (state->buffer, "else"))
        res=GRUB_PARSER_TOKEN_ELSE;
      else if (! grub_strcmp (state->buffer, "then"))
        res=GRUB_PARSER_TOKEN_THEN;
      else if (! grub_strcmp (state->buffer, "fi"))
        res=GRUB_PARSER_TOKEN_FI;
    }
  if (res==GRUB_PARSER_TOKEN_TEXT)
    {
      if ((state->offset) || (! node))
        add_str();
      yylval->node = node;
    }
  else
    {
      if (node)
        grub_script_freenode(parsestate, grub_script_getnode(parsestate, node));
    }
  state->state=newstate;
  state->read_ahead=ch;
  return res;
}

void
grub_script_yyerror (struct grub_parser_param *lex __attribute__ ((unused)),
		     char const *err)
{
  grub_printf ("%s\n", err);
}
